##################################################################
#
# General permission check
#
##################################################################

$tripconf = "/tmp/tw.$hostname";

sub chkperm {
    local($line, $command, @cmdline);
    local($file) = @_;

    die "couldn't open $file\n" unless open(KIT, "<$file");
    if ($gentrip) {
	die "Can't create tripwire configuration file $tripconf"
		unless open(TRIPCONF, ">$tripconf");
    }

    $line = 0;
    until (eof(KIT)) {
    	$conditonal = 0;
	@cmdline = ();
	while (<KIT>) {
	    chop; $line++;
	    last unless (m/\\$/);
	    s///;
	    push(@cmdline, split);
	}
	s/\s*#.*//;
	push(@cmdline, split);
	next if ($#cmdline < 0);
	$command = shift(@cmdline);
	eval(&$command(@cmdline));
    }

    close TRIPCONF if ($gentrip);
    close KIT;
}

sub package {
    $package = "@_";
    $packagefound = 0;
    @searchdirs = ();
    @ownerlist = ();
    @grouplist = ();
    $bindir = "";
    $libdir = "";
    $confdir = "";

    &setwarntag("PERM: $package");
}

sub ualias {
    local($alias) = shift(@_);

    $UALIAS{$alias} = "@_";
}

sub galias {
    local($alias) = shift(@_);

    $GALIAS{$alias} = @_;
}

sub owner {
    push(@ownerlist, &expand_own(@_));
}

sub group {
    push(@grouplist, @_);
}

sub expand_own {
    local($owner, @ret);

    foreach $owner (@_) {
	if ($owner =~ m/^%/) {
		$owner =~ s///;
		push(@ret, $UALIAS{$owner});
	} else {
		push(@ret, $owner);
	}
    }
    return @ret;
}

sub expand_grp {
    local($group, @ret);

    foreach $group (@_) {
	if ($group =~ m/^%/) {
		$group =~ s///;
		push(@ret, $GALIAS{$group});
	} else {
		push(@ret, $group);
	}
    }
    return @ret;
}

sub search {
    push(@searchdirs, @_);
    $packagefound = 1;
}

sub locate {
    local($req, $what, $name, @dirlist) = @_;
    local($ct);

    @searchdirs = &findin($name, @dirlist);

    $ct = $#searchdirs;
    if ($ct < 0) {
    	&warn("$what not found") if ($req eq "REQ");
    	$packagefound = 0;
    } elsif ($ct == 0) {
    	&mesg("$what found in @searchdirs");
    	$packagefound = 1;
    } else {
    	&warn("$what found in several places: @searchdirs");
    	$packagefound = 1;
    }
}

sub file {
    local($req, $name, $owners, $groups, $maxmode, $minmode, $tflags) = @_;

    return if (!$packagefound);
    $maxmode = oct($maxmode); $minmode = oct($minmode);
    $owners = &expand_own($owners) if ($owners != "!");
    $groups = &expand_grp($groups) if ($groups != "!");

    @pathlist = &findfile($name);
    if ($#pathlist < 0) {
	return if ("$req" eq "OPT" || "$req" eq "MAY");
	&warn("Required file $name missing");
	return;
    }
    if ($gentrip) {
    	foreach $path (@pathlist) {
	    &tripgen("file", $path, $tflags);
	}
    }

    foreach $path (@pathlist) {
	&checkperms($path, $req, $owners, $groups, $maxmode, $minmode);
    }
}

sub dir {
    local($req, $name, $owners, $groups, $maxmode, $minmode, $tflags) = @_;

    return if (!$packagefound);
    $maxmode = oct($maxmode); $minmode = oct($minmode);
    $owners = &expand_own($owners) if ($owners != "!");
    $groups = &expand_grp($groups) if ($groups != "!");

    @pathlist = &finddir($name);
    if ($#pathlist < 0) {
	return if ("$req" eq "OPT" || "$req" eq "MAY");
	&warn("Required directory $name missing");
	return;
    }
    if ($gentrip) {
    	foreach $path (@pathlist) {
	    &tripgen("dir", "=$path", $tflags);
	}
    }

    foreach $path (@pathlist) {
	&checkperms($path, $req, $owners, $groups, $maxmode, $minmode);
    }
}

sub glob {
    local($req, $name, $owners, $groups, $maxmode, $minmode, $tflags) = @_;

    return if (!$packagefound);
    $maxmode = oct($maxmode); $minmode = oct($minmode);
    $owners = &expand_own($owners) if ($owners != "!");
    $groups = &expand_grp($groups) if ($groups != "!");

    @pathlist = &findglob($name);
    if ($#pathlist < 0) {
	return if ("$req" eq "OPT" || "$req" eq "MAY");
	&warn("Required file $name missing");
	return;
    }
    if ($gentrip) {
    	foreach $path (@pathlist) {
	    &tripgen("glob", "=$path", $tflags);
	}
    }

    foreach $path (@pathlist) {
	&checkperms($path, $req, $owners, $groups, $maxmode, $minmode);
    }
}

sub tree {
    local($req, $name, $owners, $groups, $maxmode, $minmode, $tflags) = @_;
    local(@pathlist, $path);

    return if (!$packagefound);
    $maxmode = oct($maxmode); $minmode = oct($minmode);
    $owners = &expand_own($owners) if ($owners != "!");
    $groups = &expand_grp($groups) if ($groups != "!");

    @pathlist = &finddir($name);
    if ($#pathlist < 0) {
	return if ("$req" eq "OPT" || "$req" eq "MAY");
	&warn("Required file $name missing");
	return;
    }
    if ($#pathlist > 0) {
	&warn("More than one tree named $name: @pathlist\n");
	return;
    }
    if ($gentrip) {
    	foreach $path (@pathlist) {
	    &tripgen("tree", $path, $tflags);
	}
    }

    &subtree("REQ", @pathlist, $owners, $groups, $maxmode, $minmode);
}

sub subtree {
    local($req, $name, $owners, $groups, $maxmode, $minmode) = @_;
    local(@pathlist, $path);

    @pathlist = &findglob("$name/*");
    return if ($#pathlist < 0);

    foreach $path (@pathlist) {
	if (-d $path) {
		&subtree($req, $path, $owners, $groups, $maxmode, $minmode);
	} else {
		&checkperms($path, $req, $owners, $groups, $maxmode, $minmode);
	}
    }
}

sub checkperms {
    local($path, $req, $owners, $groups, $maxmode, $minmode) = @_;
    local($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size) = stat($path);

    return if ($DONE{$dev,$ino});

    $conditional = ($req eq "MAY");
    $failed = 0;

    @owners = ($owners ne "!") ? $owners : @ownerlist;
    @groups = ($groups ne "!") ? $groups : @grouplist;
    &error("empty owner list and no owner given") unless(@owners);
    &error("empty group list and no group given") unless(@groups);

    $groups = @grouplist if ($groups eq "GROUP");
    $mode   = $mode & oct(6777);
    $fowner = $LOGIN{$uid};
    $fgroup = $GROUP{$gid};

    &wperm ("Bad owner $fowner for $path; permitted: @owners")
	    unless grep(/$fowner/, @owners);
    &wperm ("Bad group $fgroup for $path; permitted: @groups")
	    unless grep(/$fgroup/, @groups);

    if (($mode | $maxmode) != $maxmode) {
	$bad = ($mode | $maxmode) ^ $maxmode;
	&wperm ("Bad setuid bit for $path")   if ($bad & $SUID);
	&wperm ("Bad setgid bit for $path")   if ($bad & $SGID);
	&wperm ("Bad r permission for $path") if ($bad & $RMODE);
	&wperm ("Bad w permission for $path") if ($bad & $WMODE);
	&wperm ("Bad x permission for $path") if ($bad & $XMODE);
    }
    if (($mode & $minmode) != $minmode) {
	$bad = ($mode ^ $minmode) & $minmode;
	&wperm ("Missing setuid bit for $path")   if ($bad & $SUID);
	&wperm ("Missing setgid bit for $path")   if ($bad & $SGID);
	&wperm ("Missing r permission for $path") if ($bad & $RMODE);
	&wperm ("Missing w permission for $path") if ($bad & $WMODE);
	&wperm ("Missing x permission for $path") if ($bad & $XMODE);
    }
    push(@SUIDBINS, $path) if ($mode & $SUID);
    push(@SGIDBINS, $path) if ($mode & $SGID);

    $DONE{$dev,$ino} = !($failed && $conditional);
}

sub findfile {
    local($file) = @_;
    local(@ret) = ();

    if ($file =~ m:^/:) {
    	return $file if (-f $file);
	return @ret;
    }
    foreach $dir (@searchdirs) {
	push(@ret, "$dir/$file") if -f "$dir/$file";
    }
    return @ret;
}

sub finddir {
    local($file) = @_;
    local(@ret) = ();

    if ($file =~ m:^/:) {
    	return $file if (-d $file);
	return @ret;
    }
    foreach $dir (@searchdirs) {
	push(@ret, "$dir/$file") if -d "$dir/$file";
    }
    return @ret;
}

sub findglob {
    local($file) = @_;
    local(@ret) = ();

    return <${file}> if ($file =~ m:^/:);
    foreach $dir (@searchdirs) {
	push(@ret, <${dir}/${file}>);
    }
    return @ret;
}

sub findin {
    local($file, @dirlist) = @_;
    local(@ret) = ();

    if ($file =~ m:^/:) {
    	return $file if (-f $file);
	return @ret;
    }
    foreach $dir (@dirlist) {
	push(@ret, $dir) if -f "$dir/$file";
    }
    return @ret;
};

sub checklist {
    local($what) = shift;

    foreach $item (@_) {
	return 1 if ($item eq $what);
    }
    return 0;
}

sub tripgen {
    local($caller, $path, $tflags) = @_;

    if ($tflags ne "") {
	print TRIPCONF "$path $tflags # $caller\n";
    } else {
	print TRIPCONF "$path R # $caller\n";
    }
}

sub wperm {

    $failed = 1;
    &warn(@_) unless ($conditional);
}

sub error {
    print STDERR "\nSYNTAX ERROR\n";
    print STDERR ">>> Error line $line: @_\n";
    $errors++;
}

1;
